@@ -4,6 +4,7 @@ import csv
 import hashlib
 import io
 import os
+import posixpath
 import shutil
 import stat
 import sys
@@ -20,14 +21,19 @@ from pdm.backend._vendor.packaging.utils import canonicalize_name
 from pdm.backend.base import Builder
 from pdm.backend.hooks import Context
 from pdm.backend.hooks.setuptools import SetuptoolsBuildHook
+from pdm.backend.structures import FileMap
 from pdm.backend.utils import (
-    expand_vars,
     get_abi_tag,
     get_platform,
+    normalize_file_permissions,
     safe_version,
     to_filename,
 )
 
+SCHEME_NAMES = frozenset(
+    ["purelib", "platlib", "include", "platinclude", "scripts", "data"]
+)
+
 if sys.version_info < (3, 8):
     from importlib_metadata import version as get_version
 else:
@@ -72,6 +78,13 @@ class WheelBuilder(Builder):
         super().__init__(location, config_settings)
         self.__tag: str | None = None
 
+    def scheme_path(self, name: str, relative: str) -> str:
+        if name not in SCHEME_NAMES:
+            raise ValueError(
+                f"Unknown scheme name {name!r}, must be one of {SCHEME_NAMES}"
+            )
+        return f"{self.name_version}.data/{name}/{relative}"
+
     def _get_platform_tags(self) -> tuple[str | None, str | None, str | None]:
         python_tag: str | None = None
         py_limited_api: str | None = None
@@ -86,44 +99,57 @@ class WheelBuilder(Builder):
             plat_name = self.config_settings["--plat-name"]
         return python_tag, py_limited_api, plat_name
 
-    def _fix_dependencies(self) -> None:
-        """Fix the dependencies and remove dynamic variables from the metadata"""
-        metadata = self.config.metadata
-        root = self.location.as_posix()
-        if metadata.get("dependencies"):
-            metadata["dependencies"] = [
-                expand_vars(dep, root) for dep in metadata["dependencies"]
-            ]
-        if metadata.get("optional-dependencies"):
-            for name, deps in metadata["optional-dependencies"].items():
-                metadata["optional-dependencies"][name] = [
-                    expand_vars(dep, root) for dep in deps
-                ]
-
-    def initialize(self, context: Context) -> None:
-        self._fix_dependencies()
-        return super().initialize(context)
-
     def prepare_metadata(self, metadata_directory: str) -> Path:
         """Write the dist-info files under the given directory"""
         context = self.build_context(Path(metadata_directory))
         self.initialize(context)
         return self._write_dist_info(Path(metadata_directory))
 
-    def get_files(self, context: Context) -> Iterable[tuple[str, Path]]:
+    def _collect_files(self, context: Context) -> FileMap:
         package_dir = self.config.build_config.package_dir
-        for relpath, path in super().get_files(context):
+        result = FileMap()
+
+        def clean_prefix(relpath: str) -> str:
             # remove the package-dir part from the relative paths
             if package_dir and relpath.startswith(package_dir + "/"):
                 relpath = relpath[len(package_dir) + 1 :]
-            yield relpath, path
+            return relpath
+
+        result.update(
+            (clean_prefix(k), v) for k, v in super()._collect_files(context).items()
+        )
+        return result
+
+    def get_files(self, context: Context) -> Iterable[tuple[str, Path]]:
+        yield from super().get_files(context)
         yield from self._get_metadata_files(context)
+        yield from self._get_wheel_data(context)
+
+    def _get_wheel_data(self, context: Context) -> Iterable[tuple[str, Path]]:
+        for name, paths in context.config.build_config.wheel_data.items():
+            for path in paths:
+                relative_to: str | None = None
+                if not isinstance(path, str):
+                    relative_to = path.get("relative-to")
+                    path = path["path"]
+                for child in context.expand_paths(path):
+                    relpath = (
+                        child.relative_to(relative_to).as_posix()
+                        if relative_to
+                        else child.name
+                    )
+                    yield self.scheme_path(name, relpath), child
 
     def build_artifact(
         self, context: Context, files: Iterable[tuple[str, Path]]
     ) -> Path:
         records: list[RecordEntry] = []
-        with tempfile.NamedTemporaryFile(suffix=".whl", delete=False) as fp:
+        fd, temp_name = tempfile.mkstemp(suffix=".whl")
+        st_mode = os.stat(temp_name).st_mode
+        new_mode = normalize_file_permissions(st_mode)
+        os.chmod(temp_name, new_mode)
+
+        with os.fdopen(fd, "w+b") as fp:
             with zipfile.ZipFile(fp, "w", compression=zipfile.ZIP_DEFLATED) as zf:
                 for rel_path, full_path in files:
                     records.append(self._add_file_to_zip(zf, rel_path, full_path))
@@ -132,7 +158,7 @@ class WheelBuilder(Builder):
         target = context.dist_dir / f"{self.name_version}-{self.tag}.whl"
         if target.exists():
             target.unlink()
-        shutil.move(fp.name, target)
+        shutil.move(temp_name, target)
         return target
 
     @property
@@ -184,7 +210,7 @@ class WheelBuilder(Builder):
     def _write_dist_info(self, parent: Path) -> Path:
         """write the dist-info directory and return the path to it"""
         dist_info = parent / self.dist_info_name
-        dist_info.mkdir(0o700, parents=True, exist_ok=True)
+        dist_info.mkdir(0o700, exist_ok=True)
         meta = self.config.metadata
         entry_points = meta.entry_points
         if entry_points:
@@ -195,7 +221,7 @@ class WheelBuilder(Builder):
             self._write_wheel_file(f, is_purelib=self.config.build_config.is_purelib)
 
         with _open_for_write(dist_info / "METADATA") as f:
-            f.write(self.format_pkginfo())
+            f.write(self.config.to_coremetadata())
 
         for file in self.find_license_files():
             target = dist_info / "licenses" / file
@@ -208,8 +234,9 @@ class WheelBuilder(Builder):
     ) -> RecordEntry:
         self._show_add_file(rel_path, full_path)
         zi = zipfile.ZipInfo(rel_path, ZIPINFO_DATE_TIME)
-        st_mode = os.stat(full_path).st_mode
-        zi.external_attr = (st_mode & 0xFFFF) << 16  # Unix attributes
+        st_mode = full_path.stat().st_mode
+        new_mode = normalize_file_permissions(st_mode)
+        zi.external_attr = (new_mode & 0xFFFF) << 16  # Unix attributes
 
         if stat.S_ISDIR(st_mode):
             zi.external_attr |= 0x10  # MS-DOS directory flag
@@ -224,11 +251,9 @@ class WheelBuilder(Builder):
         zf: zipfile.ZipFile, zi: zipfile.ZipInfo, src: IO[bytes]
     ) -> str:
         hashsum = hashlib.sha256()
-        for buf in iter(lambda: src.read(2**16), b""):
-            hashsum.update(buf)
-
-        src.seek(0)
-        zf.writestr(zi, src.read(), compress_type=zipfile.ZIP_DEFLATED)
+        data = src.read()
+        hashsum.update(data)
+        zf.writestr(zi, data, compress_type=zipfile.ZIP_DEFLATED)
         return urlsafe_b64encode(hashsum.digest()).decode("ascii").rstrip("=")
 
     def _write_record(self, zf: zipfile.ZipFile, records: list[RecordEntry]) -> None:
@@ -269,13 +294,19 @@ class WheelBuilder(Builder):
     def _get_metadata_files(self, context: Context) -> Iterable[tuple[str, Path]]:
         """Generate the metadata files for the wheel."""
         if context.kwargs.get("metadata_directory"):
-            return self._iter_files_in_directory(context.kwargs["metadata_directory"])
+            return self._iter_metadata_files(context.kwargs["metadata_directory"])
         else:
-            dist_info = self._write_dist_info(context.build_dir)
-            return self._iter_files_in_directory(str(dist_info))
+            dist_info = self._write_dist_info(context.ensure_build_dir())
+            return self._iter_metadata_files(str(dist_info))
 
-    def _iter_files_in_directory(self, path: str) -> Iterable[tuple[str, Path]]:
+    def _iter_metadata_files(self, path: str) -> Iterable[tuple[str, Path]]:
+        dist_info_name = self.dist_info_name
         for root, _, files in os.walk(path):
-            relroot = os.path.relpath(root, os.path.dirname(path))
             for file in files:
-                yield (Path(relroot, file).as_posix(), Path(root) / file)
+                # the relative path is concated with the dist-info name
+                # so that the dist info name is always consistent with the current build
+                # e.g. <path>/METADATA -> <name>-<version>.dist-info/METADATA
+                relpath = posixpath.join(
+                    dist_info_name, Path(root, file).relative_to(path).as_posix()
+                )
+                yield relpath, Path(root, file)
